﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Microsoft.Kinect;
using System.IO;
using Microsoft.Speech.AudioFormat;
using Microsoft.Speech.Recognition;
using System.Windows.Threading;

namespace ClothingSizeIdentificationUsingKinect
{
    /// <summary>
    /// Interaction logic for AppWindow.xaml
    /// </summary>
    public partial class AppWindow : Window
    {
        public static readonly DependencyProperty KinectSensorProperty =
            DependencyProperty.Register(
                "KinectSensor",
                typeof(KinectSensor),
                typeof(AppWindow),
                new PropertyMetadata(null));

        #region Member Variables
        /// <summary>
        /// Active Kinect sensor
        /// </summary>
        private KinectSensor sensor;

        /// <summary>
        /// Bitmap that will hold color information
        /// </summary>
        private WriteableBitmap colorBitmap;

        /// <summary>
        /// Intermediate storage for the color data received from the camera
        /// </summary>
        private byte[] colorPixels;

        List<Button> buttons;
        static Button selected;

        float handX;
        float handY;
        private Skeleton[] Skeletons;

        /// <summary>
        /// Thickness of drawn joint lines
        /// </summary>
        private const double JointThickness = 5;

        /// <summary>
        /// Thickness of body center ellipse
        /// </summary>
        private const double BodyCenterThickness = 10;

        /// <summary>
        /// Thickness of clip edge rectangles
        /// </summary>
        private const double ClipBoundsThickness = 10;

        /// <summary>
        /// Brush used to draw skeleton center point
        /// </summary>
        private readonly Brush centerPointBrush = Brushes.Blue;

        /// <summary>
        /// Brush used for drawing joints that are currently tracked
        /// </summary>
        private readonly Brush trackedJointBrush = new SolidColorBrush(Color.FromArgb(255, 68, 192, 68));

        /// <summary>
        /// Brush used for drawing joints that are currently inferred
        /// </summary>        
        private readonly Brush inferredJointBrush = Brushes.Yellow;

        /// <summary>
        /// Pen used for drawing bones that are currently tracked
        /// </summary>
        private readonly Pen trackedBonePen = new Pen(Brushes.Green, 6);

        /// <summary>
        /// Pen used for drawing bones that are currently inferred
        /// </summary>        
        private readonly Pen inferredBonePen = new Pen(Brushes.Gray, 1);
        /// <summary>
        /// Drawing group for skeleton rendering output
        /// </summary>
        private DrawingGroup drawingGroup;

        /// <summary>
        /// Drawing image that we will display
        /// </summary>
        private DrawingImage imageSource;

        private bool running = false;
        private bool phaseOne = false;
        private bool phaseTwo = false;
        private bool isMale = false;
        private bool isFemale = false;

        private short[] _DepthPixelData;

        private const int WaveImageWidth = 500;
        private const int WaveImageHeight = 100;

        private readonly double[] energyBuffer = new double[WaveImageWidth];

        private DispatcherTimer readyTimer;
        private EnergyCalculatingPassThroughStream stream;
        private SpeechRecognitionEngine speechRecognizer;

        /// <summary>
        /// List of values ​​obtained when measuring
        /// </summary>
        List<double> ltUserHeight = new List<double>();
        List<double> ltShoulderWidth = new List<double>();
        List<double> ltShoulderHeight = new List<double>();
        List<double> ltLengthOfPants = new List<double>();
        List<double> ltWidthOfHip = new List<double>();
        List<double> ltLengthOfArm = new List<double>();
        List<double> ltUserSideWidth = new List<double>();
        #endregion Member Variables

        public AppWindow()
        {
            InitializeComponent();
        }

        #region Kinect sensor
        public KinectSensor KinectSensor
        {
            get { return (KinectSensor)GetValue(KinectSensorProperty); }
            set { SetValue(KinectSensorProperty, value); }
        }

        private void DiscoverSensor()
        {
            KinectSensor.KinectSensors.StatusChanged += KinectSensors_StatusChanged;
            this.sensor = KinectSensor.KinectSensors.FirstOrDefault(x => x.Status == KinectStatus.Connected);
        }

        private void KinectSensors_StatusChanged(object sender, StatusChangedEventArgs e)
        {
            switch (e.Status)
            {
                case KinectStatus.Connected:
                    if (this.sensor == null)
                    {
                        // connected sensor
                        sensor = e.Sensor;
                    }
                    break;

                case KinectStatus.Disconnected:
                    if (this.sensor == e.Sensor)
                    {
                        this.sensor = null;
                        this.sensor = KinectSensor.KinectSensors.FirstOrDefault(x => x.Status == KinectStatus.Connected);

                        if (this.sensor == null)
                        {
                            // no connected device
                        }
                    }
                    break;
            }
            if (e.Status == KinectStatus.Connected)
            {
                this.sensor = e.Sensor;
            }
        }
        #endregion Kinect sensor

        #region Window loaded and closing
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            DiscoverSensor();
            CanvasGender.Visibility = Visibility.Visible;
            InitializeButtons();
            Generics.ResetHandPosition(kinectButton);
            kinectButton.Click += new RoutedEventHandler(kinectButton_click);
            
            InitializeStreams();

        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            if (null != this.sensor)
            {
                this.sensor.Stop();
                this.speechRecognizer.RecognizeAsyncCancel();
                this.speechRecognizer.RecognizeAsyncStop();
                Application.Current.Shutdown();
            }
        }
        #endregion Window loaded and closing        

        private void InitializeStreams()
        {
            // Create the drawing group we'll use for drawing
            this.drawingGroup = new DrawingGroup();

            // Create an image source that we can use in our image control
            this.imageSource = new DrawingImage(this.drawingGroup);

            // Display the drawing using our image control
            SkeletonImageMain.Source = this.imageSource;

            var parameters = new TransformSmoothParameters
            {
                Smoothing = 0.3f,
                Correction = 0.0f,
                Prediction = 0.0f,
                JitterRadius = 1.0f,
                MaxDeviationRadius = 0.5f
            };

            this.sensor.SkeletonStream.Enable(parameters);
            this.sensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30);
            this.colorPixels = new byte[this.sensor.ColorStream.FramePixelDataLength];
            this.colorBitmap = new WriteableBitmap(this.sensor.ColorStream.FrameWidth, this.sensor.ColorStream.FrameHeight, 96.0, 96.0, PixelFormats.Bgr32, null);
            this.ColorImage.Source = this.colorBitmap;
            this.ColorImageMain.Source = this.colorBitmap;
            this.sensor.ColorFrameReady += this.Sensor_ColorFrameReady;
            this.sensor.SkeletonFrameReady += this.Sensor_SkeletonFrameReady;

            this.Skeletons = new Skeleton[this.sensor.SkeletonStream.FrameSkeletonArrayLength];

            this.speechRecognizer = this.CreateSpeechRecognizer();

            InitializeDepthStream();

            try
            {
                this.sensor.Start();

            }
            catch (IOException)
            {
                this.sensor = null;
            }

            if (this.speechRecognizer != null && sensor != null)
            {
                // NOTE: Need to wait 4 seconds for device to be ready to stream audio right after initialization
                this.readyTimer = new DispatcherTimer();
                this.readyTimer.Tick += this.ReadyTimerTick;
                this.readyTimer.Interval = new TimeSpan(0, 0, 4);
                this.readyTimer.Start();

                Speaker.Content = "Inicializuji audio...";
            }
        }

        private void ReadyTimerTick(object sender, EventArgs e)
        {
            this.Start();
            Speaker.Content = "Mluvte...";
            this.readyTimer.Stop();
            this.readyTimer = null;
        }

        private void InitializeDepthStream()
        {
            this.sensor.DepthStream.Enable();

            //this.depthBitmap = new WriteableBitmap(this.sensor.DepthStream.FrameWidth, sensor.DepthStream.FrameHeight, 96, 96, PixelFormats.Bgr32, null);
            //this._DepthImageStride = this.sensor.DepthStream.FrameWidth * 4;
            this._DepthPixelData = new short[this.sensor.DepthStream.FramePixelDataLength];
            //this.DepthImageMain.Source = this.depthBitmap;

            this.sensor.DepthFrameReady += this.Sensor_DepthFrameReady;
        }

        #region All frame ready events
        /// <summary>
        /// Event handler for Kinect sensor's ColorFrameReady event
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void Sensor_ColorFrameReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame colorFrame = e.OpenColorImageFrame())
            {
                if (colorFrame != null)
                {
                    // Copy the pixel data from the image to a temporary array
                    colorFrame.CopyPixelDataTo(this.colorPixels);

                    // Write the pixel data into our bitmap
                    this.colorBitmap.WritePixels(
                        new Int32Rect(0, 0, this.colorBitmap.PixelWidth, this.colorBitmap.PixelHeight),
                        this.colorPixels,
                        this.colorBitmap.PixelWidth * sizeof(int),
                        0);
                }
            }
        }

        private void Sensor_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            Skeleton skel = null;
            using (SkeletonFrame frame = e.OpenSkeletonFrame())
            {
                if (frame != null)
                {
                    frame.CopySkeletonDataTo(this.Skeletons);
                    Skeleton skeleton = GetPrimarySkeleton(this.Skeletons);
                    skel = skeleton;

                    if (skeleton == null)
                    {
                        kinectButton.Visibility = Visibility.Collapsed;
                    }
                    else
                    {
                        Joint primaryHand = GetPrimaryHand(skeleton);
                        TrackHand(primaryHand);
                    }

                }
            }

            using (DrawingContext dc = this.drawingGroup.Open())
            {
                // Draw a transparent background to set the render size
                dc.DrawRectangle(Brushes.Transparent, null, new Rect(0.0, 0.0, this.colorBitmap.Width, this.colorBitmap.Height));

                if (skel != null)
                {
                    RenderClippedEdges(skel, dc);

                    if (skel.TrackingState == SkeletonTrackingState.Tracked)
                    {
                        this.DrawBonesAndJoints(skel, dc);

                        SkeletonData mySkelData = new SkeletonData(skel);

                        // Check user distance
                        int userDistance = mySkelData.DistanceFromSensor();
                        this.checkUserDistance(userDistance);

                        // Check user stance
                        int stance = mySkelData.UserStance();
                        this.checkUserStance(stance);

                        if (userDistance == 1 && stance == 1)
                        {
                            Orders.Content = "připraveno";
                            Orders.Foreground = Brushes.Green;
                        }
                        else
                        {
                            this.running = false;
                        }

                        // Start measuring - phase one
                        if (this.running && !this.phaseOne)
                        {
                            ltUserHeight.Add(mySkelData.UserHeight());
                            ltShoulderWidth.Add(mySkelData.ShoulderWidth());
                            ltShoulderHeight.Add(mySkelData.ShoulderHeight());
                            ltLengthOfPants.Add(mySkelData.LengthOfPants());
                            ltWidthOfHip.Add(mySkelData.WidthOfHip());
                            ltLengthOfArm.Add(mySkelData.LengthOfArm());
                        }
                    }
                    else if (skel.TrackingState == SkeletonTrackingState.PositionOnly)
                    {
                        dc.DrawEllipse(
                        this.centerPointBrush,
                        null,
                        this.SkeletonPointToScreen(skel.Position),
                        BodyCenterThickness,
                        BodyCenterThickness);
                    }

                }

                // Prevent drawing outside of our render area
                this.drawingGroup.ClipGeometry = new RectangleGeometry(new Rect(0.0, 0.0, this.colorBitmap.Width, this.colorBitmap.Height));
            }

        }      

        /// <summary>
        /// Event handler for Kinect sensor's DepthFrameReady event
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void Sensor_DepthFrameReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (DepthImageFrame frame = e.OpenDepthImageFrame())
            {
                if (frame != null)
                {
                    frame.CopyPixelDataTo(this._DepthPixelData);
                    CalculateUser(frame, this._DepthPixelData);
                }
            }
        }
        #endregion All frame ready events

        #region Helper methods for Skeleton measurement
        public void MeasurementResults()
        {
            CanvasMain.Visibility = Visibility.Collapsed;
            CanvasResults.Visibility = Visibility.Visible;

            var userHeight = AverageValueOfList(ltUserHeight);
            var shoulderWidth = AverageValueOfList(ltShoulderWidth);
            var shoulderHeight = AverageValueOfList(ltShoulderHeight);
            var lengthOfPants = AverageValueOfList(ltLengthOfPants);
            var widthOfHip = AverageValueOfList(ltWidthOfHip);
            var LengthOfArm = AverageValueOfList(ltLengthOfArm);
            var userSideWidth = AverageValueOfList(ltUserSideWidth);
            var waistSize = WaistSize(widthOfHip, userSideWidth);
            var userWeight = BodyWeight(userHeight, widthOfHip);

            var userWeightInt = Convert.ToInt32(userWeight);
            var indexBMI = CalculateBMI(userHeight,userWeightInt);
            string BMI = "(BMI " + indexBMI.ToString("F") + " - " + ClassifyBMI(indexBMI) + ")";


            if (this.isMale)
            {
                ClothingSize maleClothes = new MaleSize();
                string tshirt = maleClothes.TshirtSize(shoulderWidth, shoulderHeight);
                string tshirtMeasurement = "(šířka ramen " + shoulderWidth.ToString("F") + " cm, " + "délka trika " + shoulderHeight.ToString("F") + " cm)";

                string pants = maleClothes.PantsSize(lengthOfPants, waistSize);
                string pantsMeasurement = "(obvod pasu " + waistSize.ToString("F") + " cm, " + "délka nohavice " + lengthOfPants.ToString("F") + " cm)";

                TshirtSize.Text = "tričko " + tshirt + " " + tshirtMeasurement;
                PantsSize.Text = "kalhoty " + pants + " " + pantsMeasurement;
            }

            if (this.isFemale)
            {
                ClothingSize femaleClothes = new FemaleSize();
                string tshirt = femaleClothes.TshirtSize(shoulderWidth, shoulderHeight);
                TshirtSize.Text = "tričko " + tshirt;
            }

            UserBodyCharacter.Text = BodyHeight(userHeight) + ", " + BodyWidth(userHeight,widthOfHip);
            UserHeight.Text = "výška " + userHeight.ToString("F") + " cm";
            UserHipWidth.Text = "šířka v pase " + widthOfHip.ToString("F") + " cm";
            UserWeight.Text = "odhadovaná hmotnost " + userWeight + " kg " + BMI;    
        }

        public void checkUserStance(int value)
        {
            switch (value)
            {
                case 1:
                    Stance.Content = "čelně";
                    break;
                case 2:
                    Stance.Content = "levým bokem";
                    break;
                case 3:
                    Stance.Content = "pravým bokem";
                    break;
                default:
                    Stance.Content = "postavte se před senzor";
                    break;
            }
        }

        public void checkUserDistance(int value)
        {
            switch (value)
            {
                case 1:
                    Orders.Content = "stojíš správně";
                    break;
                case 2:
                    Orders.Content = "jdi dál od senzoru";
                    break;
                case 3:
                    Orders.Content = "jdi blíže k senzoru";
                    break;
                default:
                    Orders.Content = "postav se před senzor";
                    break;
            }
        }

        public double AverageValueOfList(List<double> list)
        {
            double sum = 0;
            foreach (double item in list)
            {
                sum += item;
            }

            return sum / list.Count;
        }

        public string BodyHeight(double value)
        {
            if (value > 180)
            {
                return "vysoká";
            }
            else if (value > 160) 
            {
                return "střední";
            }
            else
            {
                return "malá";
            }
        }

        public string BodyWidth(double height, double width)
        {
            string body = "";
            if (height > 195)
            {
                if (width < 38)
                {
                    body = "štíhlá";
                }
                else if (width < 39)
                {
                    body = "střední";
                }
                else
                {
                    body = "obézní"; 
                }
                return body;

            }
            else if (height > 185 && height < 195)
            {
                if (width < 37)
                {
                    body = "štíhlá";
                }
                else if (width < 38)
                {
                    body = "střední";
                }
                else
                {
                    body = "obézní";
                }
                return body;
            }
            else if (height > 175 && height < 185)
            {
                if (width < 36)
                {
                    body = "štíhlá";
                }
                else if (width < 37)
                {
                    body = "střední";
                }
                else
                {
                    body = "obézní"; 
                }
                return body;
            }
            else if (height > 165 && height < 175)
            {
                if (width < 35)
                {
                    body = "štíhlá";
                }
                else if (width < 36)
                {
                    body = "střední";
                }
                else
                {
                    body = "obézní"; 
                }
                return body;
            }
            else if (height > 155 && height < 165)
            {
                if (width < 34)
                {
                    body = "štíhlá";
                }
                else if (width < 35)
                {
                    body = "střední";
                }
                else
                {
                    body = "obézní";
                }
                return body;
            }
            else
            {
                return "střední*"; 
            }
        }

        public double WaistSize(double frontWidth, double sideWidth)
        {
            return (frontWidth + sideWidth) * 2 * 0.70;
        }

        public string BodyWeight(double height, double width)
        {
            string weight = "";
            if (height > 195)
            {
                if (width < 38) 
                {
                    weight =  "90";
                }
                else if (width < 39)
                {
                    weight = "100";
                }
                else 
                {
                    weight = "110"; // a více
                }
                return weight;
                
            }
            else if (height > 185 && height < 195)
            {
                if (width < 37)
                {
                    weight = "80";
                }
                else if (width < 38)
                {
                    weight = "90";
                }
                else if (width < 39)
                {
                    weight = "100";
                }
                else
                {
                    weight = "110"; // a vice
                }
                return weight;
            }
            else if (height > 175 && height < 185)
            {
                if (width < 36)
                {
                    weight = "70";
                }
                else if (width < 37)
                {
                    weight = "80";
                }
                else
                {
                    weight = "90";  // a vice
                }
                return weight;
            }
            else if (height > 165 && height < 175)
            {
                if (width < 35)
                {
                    weight = "60";
                }
                else if (width < 36)
                {
                    weight = "70";
                }
                else
                {
                    weight = "80";  // a vice
                }
                return weight;
            }
            else if (height > 155 && height < 165)
            {
                if (width < 34)
                {
                    weight = "50";
                }
                else if (width < 35)
                {
                    weight = "60";
                }
                else
                {
                    weight = "70";  // a vice
                }
                return weight;
            }
            else
            {
                return "40";  // a mene
            }
        }

        public double CalculateBMI(double height, double weight)
        {
            return weight / Math.Pow(height/100,2);
        }

        public string ClassifyBMI(double bmi)
        {
            if (bmi > 40)
            {
                return "obezita třetího stupně";
            }
            else if (bmi > 35)
            {
                return "obezita druhého stupně";
            }
            else if (bmi > 30)
            {
                return "obezita prvního  stupně";
            }
            else if (bmi > 25)
            {
                return "nadváha";
            }
            else if (bmi > 18.5)
            {
                return "optimální váha";
            }
            else
            {
                return "podváha";
            }       
        }
        #endregion Helper methods for Skeleton measurement

        #region Helper methods for DepthImage processing
        private void CalculateUser(DepthImageFrame depthFrame, short[] pixelData)
        {
            int depth;
            int userIndex;
            int pixelIndex;
            int bytesPerPixel = depthFrame.BytesPerPixel;
            DepthData[] user = new DepthData[1];


            //First pass - Calculate stats from the pixel data
            for (int row = 0; row < depthFrame.Height; row++)
            {
                for (int col = 0; col < depthFrame.Width; col++)
                {
                    pixelIndex = col + (row * depthFrame.Width);
                    depth = pixelData[pixelIndex] >> DepthImageFrame.PlayerIndexBitmaskWidth;

                    if (depth != 0)
                    {
                        userIndex = (pixelData[pixelIndex] & DepthImageFrame.PlayerIndexBitmask) - 1;
          
                        if (userIndex > -1)
                        {
                            try
                            {                           
                                if (user[userIndex] == null)
                                {
                                    user[userIndex] = new DepthData(userIndex + 1, depthFrame.Width, depthFrame.Height);

                                }

                                user[userIndex].UpdateData(col, row, depth);

                                // start measuring for phase two
                                if (this.phaseOne && this.running)
                                {
                                    double userWidth = user[userIndex].UserWidth();
                                    ltUserSideWidth.Add(userWidth);
                                }

                            }
                            catch (IndexOutOfRangeException)
                            {
                                // nastavit hodnotu obvodu pasu cca 90
                                return;                                
                            }

                        }

                    }
                }
            }

        }
        #endregion Helper methods for depth image processing

        #region Skeleton drawing logic
        //get the skeleton closest to the Kinect sensor
        private static Skeleton GetPrimarySkeleton(Skeleton[] skeletons)
        {
            Skeleton skeleton = null;
            if (skeletons != null)
            {
                for (int i = 0; i < skeletons.Length; i++)
                {
                    if (skeletons[i].TrackingState == SkeletonTrackingState.Tracked)
                    {
                        if (skeleton == null)
                        {
                            skeleton = skeletons[i];
                        }
                        else
                        {
                            if (skeleton.Position.Z > skeletons[i].Position.Z)
                            {
                                skeleton = skeletons[i];
                            }
                        }
                    }
                }
            }
            return skeleton;
        }

        /// <summary>
        /// Draws indicators to show which edges are clipping skeleton data
        /// </summary>
        /// <param name="skeleton">skeleton to draw clipping information for</param>
        /// <param name="drawingContext">drawing context to draw to</param>
        private void RenderClippedEdges(Skeleton skeleton, DrawingContext drawingContext)
        {
            SkeletonOutLabel.Visibility = Visibility.Collapsed;

            string legs = null;
            string head = null;
            string leftHand = null;
            string rightHand = null;

            if (skeleton.ClippedEdges.HasFlag(FrameEdges.Bottom))
            {
                drawingContext.DrawRectangle(
                    Brushes.Red,
                    null,
                    new Rect(0, this.colorBitmap.Height - ClipBoundsThickness, this.colorBitmap.Width, ClipBoundsThickness));
                SkeletonOutLabel.Visibility = Visibility.Visible;
                legs = "nohy";
                this.running = false;
            }

            if (skeleton.ClippedEdges.HasFlag(FrameEdges.Top))
            {
                drawingContext.DrawRectangle(
                    Brushes.Red,
                    null,
                    new Rect(0, 0, this.colorBitmap.Width, ClipBoundsThickness));
                SkeletonOutLabel.Visibility = Visibility.Visible;
                head = "hlava";
                this.running = false;
            }

            if (skeleton.ClippedEdges.HasFlag(FrameEdges.Left))
            {
                drawingContext.DrawRectangle(
                    Brushes.Red,
                    null,
                    new Rect(0, 0, ClipBoundsThickness, this.colorBitmap.Height));
                SkeletonOutLabel.Visibility = Visibility.Visible;
                leftHand = "levá ruka";
                this.running = false;
            }

            if (skeleton.ClippedEdges.HasFlag(FrameEdges.Right))
            {
                drawingContext.DrawRectangle(
                    Brushes.Red,
                    null,
                    new Rect(this.colorBitmap.Width - ClipBoundsThickness, 0, ClipBoundsThickness, this.colorBitmap.Height));
                SkeletonOutLabel.Visibility = Visibility.Visible;
                rightHand = "pravá ruka";
                this.running = false;
            }

            legs = legs != null ? legs : "";
            head = head != null ? ", " + head : "";
            leftHand = leftHand != null ? ", " + leftHand : "";
            rightHand = rightHand != null ? ", " + rightHand : "";
            SkeletonOut.Content = legs + head + leftHand + rightHand;
        }

        /// <summary>
        /// Draws a skeleton's bones and joints
        /// </summary>
        /// <param name="skeleton">skeleton to draw</param>
        /// <param name="drawingContext">drawing context to draw to</param>
        private void DrawBonesAndJoints(Skeleton skeleton, DrawingContext drawingContext)
        {
            // Render Torso
            this.DrawBone(skeleton, drawingContext, JointType.Head, JointType.ShoulderCenter);
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderCenter, JointType.ShoulderLeft);
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderCenter, JointType.ShoulderRight);
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderCenter, JointType.Spine);
            this.DrawBone(skeleton, drawingContext, JointType.Spine, JointType.HipCenter);
            this.DrawBone(skeleton, drawingContext, JointType.HipCenter, JointType.HipLeft);
            this.DrawBone(skeleton, drawingContext, JointType.HipCenter, JointType.HipRight);

            // Left Arm
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderLeft, JointType.ElbowLeft);
            this.DrawBone(skeleton, drawingContext, JointType.ElbowLeft, JointType.WristLeft);
            this.DrawBone(skeleton, drawingContext, JointType.WristLeft, JointType.HandLeft);

            // Right Arm
            this.DrawBone(skeleton, drawingContext, JointType.ShoulderRight, JointType.ElbowRight);
            this.DrawBone(skeleton, drawingContext, JointType.ElbowRight, JointType.WristRight);
            this.DrawBone(skeleton, drawingContext, JointType.WristRight, JointType.HandRight);

            // Left Leg
            this.DrawBone(skeleton, drawingContext, JointType.HipLeft, JointType.KneeLeft);
            this.DrawBone(skeleton, drawingContext, JointType.KneeLeft, JointType.AnkleLeft);
            this.DrawBone(skeleton, drawingContext, JointType.AnkleLeft, JointType.FootLeft);

            // Right Leg
            this.DrawBone(skeleton, drawingContext, JointType.HipRight, JointType.KneeRight);
            this.DrawBone(skeleton, drawingContext, JointType.KneeRight, JointType.AnkleRight);
            this.DrawBone(skeleton, drawingContext, JointType.AnkleRight, JointType.FootRight);

            // Render Joints
            foreach (Joint joint in skeleton.Joints)
            {
                Brush drawBrush = null;

                if (joint.TrackingState == JointTrackingState.Tracked)
                {
                    drawBrush = this.trackedJointBrush;
                }
                else if (joint.TrackingState == JointTrackingState.Inferred)
                {
                    drawBrush = this.inferredJointBrush;
                }

                if (drawBrush != null)
                {
                    drawingContext.DrawEllipse(drawBrush, null, this.SkeletonPointToScreen(joint.Position), JointThickness, JointThickness);
                }
            }
        }

        /// <summary>
        /// Maps a SkeletonPoint to lie within our render space and converts to Point
        /// </summary>
        /// <param name="skelpoint">point to map</param>
        /// <returns>mapped point</returns>
        private Point SkeletonPointToScreen(SkeletonPoint skelpoint)
        {
            // Convert point to depth space.  
            // We are not using depth directly, but we do want the points in our 640x480 output resolution.
            DepthImagePoint depthPoint = this.sensor.CoordinateMapper.MapSkeletonPointToDepthPoint(skelpoint, DepthImageFormat.Resolution640x480Fps30);
            return new Point(depthPoint.X, depthPoint.Y);
        }

        /// <summary>
        /// Draws a bone line between two joints
        /// </summary>
        /// <param name="skeleton">skeleton to draw bones from</param>
        /// <param name="drawingContext">drawing context to draw to</param>
        /// <param name="jointType0">joint to start drawing from</param>
        /// <param name="jointType1">joint to end drawing at</param>
        private void DrawBone(Skeleton skeleton, DrawingContext drawingContext, JointType jointType0, JointType jointType1)
        {
            Joint joint0 = skeleton.Joints[jointType0];
            Joint joint1 = skeleton.Joints[jointType1];

            // If we can't find either of these joints, exit
            if (joint0.TrackingState == JointTrackingState.NotTracked ||
                joint1.TrackingState == JointTrackingState.NotTracked)
            {
                return;
            }

            // Don't draw if both points are inferred
            if (joint0.TrackingState == JointTrackingState.Inferred &&
                joint1.TrackingState == JointTrackingState.Inferred)
            {
                return;
            }

            // We assume all drawn bones are inferred unless BOTH joints are tracked
            Pen drawPen = this.inferredBonePen;
            if (joint0.TrackingState == JointTrackingState.Tracked && joint1.TrackingState == JointTrackingState.Tracked)
            {
                drawPen = this.trackedBonePen;
            }

            drawingContext.DrawLine(drawPen, this.SkeletonPointToScreen(joint0.Position), this.SkeletonPointToScreen(joint1.Position));
        }
        #endregion Skeleton drawing logic

        #region Hand tracking and hover button
        private void InitializeButtons()
        {
            this.buttons = new List<Button> { MensButton, WomensButton };
        }

        //track and display hand
        private void TrackHand(Joint hand)
        {
            if (hand.TrackingState == JointTrackingState.NotTracked)
            {
                kinectButton.Visibility = Visibility.Collapsed;
            }
            else
            {
                kinectButton.Visibility = Visibility.Visible;

                DepthImagePoint point = this.sensor.CoordinateMapper.MapSkeletonPointToDepthPoint(hand.Position, DepthImageFormat.Resolution640x480Fps30);
                handX = (int)((point.X * LayoutRoot.ActualWidth / this.sensor.DepthStream.FrameWidth) -
                    (kinectButton.ActualWidth / 2.0));
                handY = (int)((point.Y * LayoutRoot.ActualHeight / this.sensor.DepthStream.FrameHeight) -
                    (kinectButton.ActualHeight / 2.0));
                Canvas.SetLeft(kinectButton, handX);
                Canvas.SetTop(kinectButton, handY);

                if (isHandOver(kinectButton, buttons)) kinectButton.Hovering();
                else kinectButton.Release();

                if (hand.JointType == JointType.HandRight)
                {
                    kinectButton.ImageSource = "/ClothingSizeIdentificationUsingKinect;component/Images/RightHand.png";
                    kinectButton.ActiveImageSource = "/ClothingSizeIdentificationUsingKinect;component/Images/RightHand.png";
                }
                else
                {
                    kinectButton.ImageSource = "/ClothingSizeIdentificationUsingKinect;component/Images/LeftHand.png";
                    kinectButton.ActiveImageSource = "/ClothingSizeIdentificationUsingKinect;component/Images/LeftHand.png";
                }

            }
        }

        //detect if hand is overlapping over any button
        private bool isHandOver(FrameworkElement hand, List<Button> buttonslist)
        {

            var handTopLeft = new Point(Canvas.GetLeft(hand), Canvas.GetTop(hand));
            var handX = handTopLeft.X + hand.ActualWidth / 2;
            var handY = handTopLeft.Y + hand.ActualHeight / 2;

            foreach (Button target in buttonslist)
            {
                if (target != null)
                {
                    Point targetTopLeft = new Point(Canvas.GetLeft(target), Canvas.GetTop(target));
                    if (handX > targetTopLeft.X &&
                        handX < targetTopLeft.X + target.Width &&
                        handY > targetTopLeft.Y &&
                        handY < targetTopLeft.Y + target.Height)
                    {
                        selected = target;
                        return true;
                    }
                }
            }
            return false;
        }

        //get the hand closest to the Kinect sensor
        private static Joint GetPrimaryHand(Skeleton skeleton)
        {
            Joint primaryHand = new Joint();
            if (skeleton != null)
            {
                primaryHand = skeleton.Joints[JointType.HandLeft];
                Joint rightHand = skeleton.Joints[JointType.HandRight];
                if (rightHand.TrackingState != JointTrackingState.NotTracked)
                {
                    if (primaryHand.TrackingState == JointTrackingState.NotTracked)
                    {
                        primaryHand = rightHand;
                    }
                    else
                    {
                        if (primaryHand.Position.Z > rightHand.Position.Z)
                        {
                            primaryHand = rightHand;
                        }
                    }
                }
            }
            return primaryHand;
        }
        #endregion Hand tracking and hover button

        #region Event handlers    
        private void MensButton_Click(object sender, RoutedEventArgs e)
        {
            CanvasMain.Visibility = Visibility.Visible;
            CanvasGender.Visibility = Visibility.Collapsed;
            ColorImage.Visibility = Visibility.Collapsed;
            this.isMale = true;
        }

        private void WomensButton_Click(object sender, RoutedEventArgs e)
        {
            CanvasMain.Visibility = Visibility.Visible;
            CanvasGender.Visibility = Visibility.Collapsed;
            ColorImage.Visibility = Visibility.Collapsed;
            this.isFemale = true;
        }

        private void kinectButton_click(object sender, RoutedEventArgs e)
        {
            selected.RaiseEvent(new RoutedEventArgs(Button.ClickEvent, selected));
        }
        #endregion Event handlers

        #region Speech recognization
        private static RecognizerInfo GetKinectRecognizer()
        {
            Func<RecognizerInfo, bool> matchingFunc = r =>
            {
                string value;
                r.AdditionalInfo.TryGetValue("Kinect", out value);
                return "True".Equals(value, StringComparison.InvariantCultureIgnoreCase) && "en-US".Equals(r.Culture.Name, StringComparison.InvariantCultureIgnoreCase);
            };
            return SpeechRecognitionEngine.InstalledRecognizers().Where(matchingFunc).FirstOrDefault();
        }

        private SpeechRecognitionEngine CreateSpeechRecognizer()
        {
            RecognizerInfo ri = GetKinectRecognizer();
            if (ri == null)
            {
                MessageBox.Show(
                    @"There was a problem initializing Speech Recognition.
                    Ensure you have the Microsoft Speech SDK installed.",
                    "Failed to load Speech SDK",
                    MessageBoxButton.OK,
                    MessageBoxImage.Error);
                this.Close();
                return null;
            }

            SpeechRecognitionEngine sre;
            try
            {
                sre = new SpeechRecognitionEngine(ri.Id);
            }
            catch
            {
                MessageBox.Show(
                    @"There was a problem initializing Speech Recognition.
                    Ensure you have the Microsoft Speech SDK installed and configured.",
                    "Failed to load Speech SDK",
                    MessageBoxButton.OK,
                    MessageBoxImage.Error);
                this.Close();
                return null;
            }

            var grammar = new Choices();
            grammar.Add("man");
            grammar.Add("women");
            grammar.Add("first");
            grammar.Add("close");
            grammar.Add("show");
            grammar.Add("stop");
            grammar.Add("second");
            grammar.Add("restart");

            var gb = new GrammarBuilder { Culture = ri.Culture };
            gb.Append(grammar);

            // Create the actual Grammar instance, and then load it into the speech recognizer.
            var g = new Grammar(gb);

            sre.LoadGrammar(g);
            sre.SpeechRecognized += this.SreSpeechRecognized;


            return sre;
        }

        private void Start()
        {
            this.sensor.AudioSource.Start();
            var audioSource = this.sensor.AudioSource;
            audioSource.BeamAngleMode = BeamAngleMode.Adaptive;
            var kinectStream = audioSource.Start();
            this.stream = new EnergyCalculatingPassThroughStream(kinectStream);
            this.speechRecognizer.SetInputToAudioStream(
                this.stream, new SpeechAudioFormatInfo(EncodingFormat.Pcm, 16000, 16, 1, 32000, 2, null));
            this.speechRecognizer.RecognizeAsync(RecognizeMode.Multiple);
        }


        private void SreSpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {
            if (e.Result.Confidence < 0.7)
            {
                return;
            }

            switch (e.Result.Text.ToUpperInvariant())
            {
                case "WOMEN":
                    WomensButton.RaiseEvent(new RoutedEventArgs(Button.ClickEvent, WomensButton));
                    SpeakerLabel.Visibility = Visibility.Visible;
                    break;
                case "MAN":
                    MensButton.RaiseEvent(new RoutedEventArgs(Button.ClickEvent, MensButton));
                    SpeakerLabel.Visibility = Visibility.Visible;
                    break;
                case "FIRST":
                    this.running = true;
                    MeasurementStatus.Content = "fáze 1 probíhá";
                    MeasurementStatus.Foreground = Brushes.Blue;

                    SpeakerLabel.Visibility = Visibility.Visible;

                    this.readyTimer = new DispatcherTimer();
                    this.readyTimer.Tick += this.PhaseOneTimerTick;
                    this.readyTimer.Interval = new TimeSpan(0, 0, 5);
                    this.readyTimer.Start();

                    break;
                case "SECOND":                                        
                    this.running = true;
                    MeasurementStatus.Content = "fáze 2 probíhá";
                    MeasurementStatus.Foreground = Brushes.Blue;

                    SpeakerLabel.Visibility = Visibility.Visible;

                    this.readyTimer = new DispatcherTimer();
                    this.readyTimer.Tick += this.PhaseTwoTimerTick;
                    this.readyTimer.Interval = new TimeSpan(0, 0, 5);
                    this.readyTimer.Start();

                    break;
                case "STOP":
                    this.running = false;
                    MeasurementStatus.Content = "zastaveno";
                    MeasurementStatus.Foreground = Brushes.Red;

                    SpeakerLabel.Visibility = Visibility.Visible;
                    break;
                case "SHOW":
                    MeasurementResults();

                    SpeakerLabel.Visibility = Visibility.Visible;
                    break;
                case "RESTART":
                    NewMeasurement();
                    break;
                case "CLOSE":
                    this.Close();
                    break;
                default:

                    break;
            }

            Speaker.Content = e.Result.Text;
        }

        private class EnergyCalculatingPassThroughStream : Stream
        {
            private const int SamplesPerPixel = 10;

            private readonly double[] energy = new double[WaveImageWidth];
            private readonly object syncRoot = new object();
            private readonly Stream baseStream;

            private int index;
            private int sampleCount;
            private double avgSample;

            public EnergyCalculatingPassThroughStream(Stream stream)
            {
                this.baseStream = stream;
            }

            public override long Length
            {
                get { return this.baseStream.Length; }
            }

            public override long Position
            {
                get { return this.baseStream.Position; }
                set { this.baseStream.Position = value; }
            }

            public override bool CanRead
            {
                get { return this.baseStream.CanRead; }
            }

            public override bool CanSeek
            {
                get { return this.baseStream.CanSeek; }
            }

            public override bool CanWrite
            {
                get { return this.baseStream.CanWrite; }
            }

            public override void Flush()
            {
                this.baseStream.Flush();
            }

            public void GetEnergy(double[] energyBuffer)
            {
                lock (this.syncRoot)
                {
                    int energyIndex = this.index;
                    for (int i = 0; i < this.energy.Length; i++)
                    {
                        energyBuffer[i] = this.energy[energyIndex];
                        energyIndex++;
                        if (energyIndex >= this.energy.Length)
                        {
                            energyIndex = 0;
                        }
                    }
                }
            }

            public override int Read(byte[] buffer, int offset, int count)
            {
                int retVal = this.baseStream.Read(buffer, offset, count);
                const double A = 0.3;
                lock (this.syncRoot)
                {
                    for (int i = 0; i < retVal; i += 2)
                    {
                        short sample = BitConverter.ToInt16(buffer, i + offset);
                        this.avgSample += sample * sample;
                        this.sampleCount++;

                        if (this.sampleCount == SamplesPerPixel)
                        {
                            this.avgSample /= SamplesPerPixel;

                            this.energy[this.index] = .2 + ((this.avgSample * 11) / (int.MaxValue / 2));
                            this.energy[this.index] = this.energy[this.index] > 10 ? 10 : this.energy[this.index];

                            if (this.index > 0)
                            {
                                this.energy[this.index] = (this.energy[this.index] * A) + ((1 - A) * this.energy[this.index - 1]);
                            }

                            this.index++;
                            if (this.index >= this.energy.Length)
                            {
                                this.index = 0;
                            }

                            this.avgSample = 0;
                            this.sampleCount = 0;
                        }
                    }
                }

                return retVal;
            }

            public override long Seek(long offset, SeekOrigin origin)
            {
                return this.baseStream.Seek(offset, origin);
            }

            public override void SetLength(long value)
            {
                this.baseStream.SetLength(value);
            }

            public override void Write(byte[] buffer, int offset, int count)
            {
                this.baseStream.Write(buffer, offset, count);
            }
        }
        #endregion Speech recognization

        #region Measurement methods
        private void PhaseOneTimerTick(object sender, EventArgs e)
        {
            var dispatcherTimer = (DispatcherTimer)sender;
            dispatcherTimer.Stop();

            this.running = false;
            this.phaseOne = true;

            MeasurementStatus.Content = "následuj pokyny";
            MeasurementStatus.Foreground = Brushes.Orange;

            PhaseOneStatus.Content = "ok";
            PhaseOneStatus.Foreground = Brushes.Green;

            Orders.Content = "otočte se bokem k senzoru";
        }

        private void PhaseTwoTimerTick(object sender, EventArgs e)
        {
            var dispatcherTimer = (DispatcherTimer)sender;
            dispatcherTimer.Stop();

            this.running = false;
            this.phaseTwo = true;

            MeasurementStatus.Content = "měření úspěšné";
            MeasurementStatus.Foreground = Brushes.Green;

            PhaseTwoStatus.Content = "ok";
            PhaseTwoStatus.Foreground = Brushes.Green;

            this.readyTimer = new DispatcherTimer();
            this.readyTimer.Tick += this.MeasurementResultsTimerTick;
            this.readyTimer.Interval = new TimeSpan(0, 0, 3);
            this.readyTimer.Start();
        }

        private void MeasurementResultsTimerTick(object sender, EventArgs e)
        {
            var dispatcherTimer = (DispatcherTimer)sender;
            dispatcherTimer.Stop();

            SpeakerLabel.Visibility = Visibility.Hidden;
            Speaker.Visibility = Visibility.Hidden;

            MeasurementResults();


        }

        private void NewMeasurement()
        {
            MeasurementStatus.Content = "následuj pokyny";
            MeasurementStatus.Foreground = Brushes.Orange;
            PhaseOneStatus.Content = "nezměřeno";
            PhaseOneStatus.Foreground = Brushes.Red;
            PhaseTwoStatus.Content = "nezměřeno";
            PhaseTwoStatus.Foreground = Brushes.Red;

            ltShoulderWidth.Clear();
            ltShoulderHeight.Clear();
            ltLengthOfArm.Clear();
            ltLengthOfPants.Clear();
            ltUserHeight.Clear();
            ltUserSideWidth.Clear();
            ltWidthOfHip.Clear();

            UserBodyCharacter.Text = "";
            UserHeight.Text = "";
            UserHipWidth.Text = "";
            UserWeight.Text = "";
            TshirtSize.Text = "";
            PantsSize.Text = "";

            this.running = false;
            this.phaseOne = false;
            this.phaseTwo = false;
            this.isFemale = false;
            this.isMale = false;

            SpeakerLabel.Visibility = Visibility.Visible;
            Speaker.Visibility = Visibility.Visible;

            CanvasResults.Visibility = Visibility.Hidden;
            CanvasMain.Visibility = Visibility.Hidden;
            CanvasGender.Visibility = Visibility.Visible;
        }
        #endregion Measurement methods

    }
}
