﻿using System.Diagnostics.CodeAnalysis;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using Microsoft.Kinect;
using System.Windows.Media;
using System.IO;
using System;
using System.Threading.Tasks;
using System.Collections.Generic;

namespace ClothingSizeIdentificationUsingKinect
{
    class SkeletonData
    {
        #region Helpers
        /// <summary>
        /// Divergence of height distance measuring for head
        /// </summary>
        const double headDivergence = 0.2;

        /// <summary>
        /// Divergence of shoulder distance measuring
        /// </summary>
        const double shoulderDivergence = 0.11;

        /// <summary>
        /// Divergence of shoulder center and hip center distance measuring
        /// </summary>
        const double shoulderAndHipDivergence = 0.15;

        /// <summary>
        /// Divergence of leg distance measuring
        /// </summary>
        const double legDivergence = 0.0;

        /// <summary>
        /// Divergence of hip distance measuring
        /// </summary>
        const double hipDivergence = 0.15;

        /// <summary>
        /// Divergence of arm distance measuring
        /// </summary>
        const double armDivergence = 0.05;

        Skeleton skeleton;

        #endregion Helpers

        #region Constructor
        public SkeletonData(Skeleton skeleton)
        {
            this.skeleton = skeleton;
        }
        #endregion Constructor

        #region User position and stance
        /// <summary>
        /// Returns user stance.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <returns>Stance of the user as string.</returns>
        public int UserStance()
        {
            var shoulderLeft = this.skeleton.Joints[JointType.ShoulderLeft];
            var shoulderRight = this.skeleton.Joints[JointType.ShoulderRight];

            if (shoulderLeft.TrackingState == JointTrackingState.Inferred)
            {
                // Right side
                return 3;
            }

            if (shoulderRight.TrackingState == JointTrackingState.Inferred)
            {
                // Left side
                return 2;
            }

            // Facing
            return 1;

        }

        public int DistanceFromSensor()
        {
            var head = this.skeleton.Joints[JointType.Head];

            if (head.Position.Z > 2.9)
            {
                // Go closer
                return 3;
            }

            if (head.Position.Z < 2.7)
            {
                // Go away
                return 2;
            }

            // Standing correctly
            return 1;

        }
        #endregion User position and stance

        #region User body calculation
        /// <summary>
        /// Returns the user height.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <returns>The height of the skeleton in centimeters.</returns>
        public double UserHeight()
        {
            var head = this.skeleton.Joints[JointType.Head];
            var shoulderCenter = this.skeleton.Joints[JointType.ShoulderCenter];
            var spine = this.skeleton.Joints[JointType.Spine];
            var hipCenter = this.skeleton.Joints[JointType.HipCenter];
            var hipLeft = this.skeleton.Joints[JointType.HipLeft];
            var hipRight = this.skeleton.Joints[JointType.HipRight];
            var kneeLeft = this.skeleton.Joints[JointType.KneeLeft];
            var kneeRight = this.skeleton.Joints[JointType.KneeRight];
            var ankleLeft = this.skeleton.Joints[JointType.AnkleLeft];
            var ankleRight = this.skeleton.Joints[JointType.AnkleRight];
            var footLeft = this.skeleton.Joints[JointType.FootLeft];
            var footRight = this.skeleton.Joints[JointType.FootRight];

            // Find which leg is tracked more accurately.
            int legLeftTrackedJoints = CountOfTrackedJoints(hipLeft, kneeLeft, ankleLeft, footLeft);
            int legRightTrackedJoints = CountOfTrackedJoints(hipRight, kneeRight, ankleRight, footRight);

            double legLength = legLeftTrackedJoints > legRightTrackedJoints ? JointsLength(hipLeft, kneeLeft, ankleLeft, footLeft) : JointsLength(hipRight, kneeRight, ankleRight, footRight);

            return (JointsLength(head, shoulderCenter, spine, hipCenter) + legLength + headDivergence) * 100;
        }

        /// <summary>
        /// Returns the shoulder width.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <returns>The width between left and right shoulder in centimeters.</returns>
        public double ShoulderWidth()
        {
            var shoulderLeft = this.skeleton.Joints[JointType.ShoulderLeft];
            var shoulderCenter = this.skeleton.Joints[JointType.ShoulderCenter];
            var shoulderRight = this.skeleton.Joints[JointType.ShoulderRight];

            double shoulderWidth = JointsLength(shoulderLeft, shoulderCenter, shoulderRight) + shoulderDivergence;

            return (JointsLength(shoulderLeft, shoulderCenter, shoulderRight) + shoulderDivergence) * 100;
        }

        /// <summary>
        /// Returns width between shoulder and hip center.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <returns>The width between shoulder center and hip center in centimeters.</returns>
        public double ShoulderHeight()
        {
            var shoulderCenter = this.skeleton.Joints[JointType.ShoulderCenter];
            var hipCenter = this.skeleton.Joints[JointType.HipCenter];

            double shoulderHeight = JointsLength(shoulderCenter, hipCenter) + shoulderAndHipDivergence;

            return (JointsLength(shoulderCenter, hipCenter) + shoulderAndHipDivergence) * 100;
        }

        /// <summary>
        /// Returns length of arm.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <returns>The width between shoulder and wrist in centimeters.</returns>
        public double LengthOfArm()
        {
            var shoulderLeft = this.skeleton.Joints[JointType.ShoulderLeft];
            var shoulderRight = this.skeleton.Joints[JointType.ShoulderRight];
            var elbowLeft = this.skeleton.Joints[JointType.ElbowLeft];
            var elbowRight = this.skeleton.Joints[JointType.ElbowRight];
            var wristLeft = this.skeleton.Joints[JointType.WristLeft];
            var wristRight = this.skeleton.Joints[JointType.WristRight];

            // Find which leg is tracked more accurately.
            int armLeftTrackedJoints = CountOfTrackedJoints(shoulderLeft, elbowLeft, wristLeft);
            int armRightTrackedJoints = CountOfTrackedJoints(shoulderRight, elbowRight, wristRight);

            double armLength = armLeftTrackedJoints > armRightTrackedJoints ? JointsLength(shoulderLeft, elbowLeft, wristLeft) : JointsLength(shoulderRight, elbowRight, wristRight);

            return (armLength + armDivergence) * 100;
        }

        /// <summary>
        /// Returns length for pants.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <returns>The width between hip center and left/right in centimeters.</returns>
        public double LengthOfPants()
        {
            var hipCenter = this.skeleton.Joints[JointType.HipCenter];
            var hipLeft = this.skeleton.Joints[JointType.HipLeft];
            var hipRight = this.skeleton.Joints[JointType.HipRight];
            var kneeLeft = this.skeleton.Joints[JointType.KneeLeft];
            var kneeRight = this.skeleton.Joints[JointType.KneeRight];
            var ankleLeft = this.skeleton.Joints[JointType.AnkleLeft];
            var ankleRight = this.skeleton.Joints[JointType.AnkleRight];

            // Find which leg is tracked more accurately.
            int legLeftTrackedJoints = CountOfTrackedJoints(hipLeft, kneeLeft, ankleLeft);
            int legRightTrackedJoints = CountOfTrackedJoints(hipRight, kneeRight, ankleRight);

            double legLength = legLeftTrackedJoints > legRightTrackedJoints ? JointsLength(hipCenter, hipLeft, kneeLeft, ankleLeft) : JointsLength(hipCenter, hipRight, kneeRight, ankleRight);

            return (legLength + legDivergence) * 100;
        }

        /// <summary>
        /// Returns width of hips.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <returns>The width between left hip and right hip in centimeters.</returns>
        public double WidthOfHip()
        {
            var hipCenter = this.skeleton.Joints[JointType.HipCenter];
            var hipLeft = this.skeleton.Joints[JointType.HipLeft];
            var hipRight = this.skeleton.Joints[JointType.HipRight];

            return (JointsLength(hipLeft, hipCenter, hipRight) + hipDivergence) * 100;
        }
        #endregion User body calculation

        #region Helper method
        /// <summary>
        /// Retruns the length of the specified two skeleton joints.
        /// </summary>
        /// <param name="skeleton">The specified user skeleton.</param>
        /// <param name="jointType0">The specified user skeleton first joint.</param>
        /// <param name="jointType1">The specified user skeleton last joint.</param>
        /// <returns>The height of the skeleton in meters.</returns>
        public double JointsLength(Joint joint0, Joint joint1)
        {
            double jointsLength = Math.Sqrt(
               Math.Pow(joint0.Position.X - joint1.Position.X, 2) +
               Math.Pow(joint0.Position.Y - joint1.Position.Y, 2));

            return jointsLength;
        }

        /// <summary>
        /// Returns the length of the segments defined by the specified joints.
        /// </summary>
        /// <param name="joints">A collection of two or more joints.</param>
        /// <returns>The length of all the segments in meters.</returns>
        public double JointsLength(params Joint[] joints)
        {
            double length = 0;

            for (int index = 0; index < joints.Length - 1; index++)
            {
                length += JointsLength(joints[index], joints[index + 1]);
            }

            return length;
        }

        /// <summary>
        /// Given a collection of joints, calculates the number of the joints that are tracked accurately.
        /// </summary>
        /// <param name="joints">A collection of joints.</param>
        /// <returns>The count of the accurately tracked joints.</returns>
        public static int CountOfTrackedJoints(params Joint[] joints)
        {
            int trackedJoints = 0;

            foreach (var joint in joints)
            {
                if (joint.TrackingState == JointTrackingState.Tracked)
                {
                    trackedJoints++;
                }
            }

            return trackedJoints;
        }
        #endregion Helper method
    }
}


